from dataclasses import dataclass, field
import struct
from math import atan2, asin
from typing import Any, Dict, Tuple
from enum import Enum

class NAT_Data(Enum):
  MARKER_SET     = 0
  RIGID_BODY    = 1
  SKELETON      = 2
  FORCE_PLATE   = 3
  DEVICE        = 4
  CAMERA        = 5
  ASSET         = 6
  UNDEFINED     = -1

  @classmethod
  def _missing(cls, _: object):
    return cls.UNDEFINED

class NAT_Messages(Enum):
  # Client/server message ids
  CONNECT               = 0
  SERVER_INFO           = 1
  REQUEST               = 2
  RESPONSE              = 3
  REQUEST_MODEL_DEF     = 4
  MODEL_DEF             = 5
  REQUEST_FRAME_OF_DATA = 6
  FRAME_OF_DATA         = 7
  MESSAGE_STRING        = 8
  DISCONNECT            = 9
  KEEP_ALIVE            = 10
  UNRECOGNIZED_REQUEST  = 100
  UNDEFINED             = 999999

  @classmethod
  def _missing(cls, _: object):
    return cls.UNDEFINED

class Data:
  @classmethod
  def unpack(cls, data: bytes) -> Any:
    raise NotImplementedError("Subclasses must implement the unpack method")

@dataclass(frozen=True)
class Position(Data):
  x: float
  y: float
  z: float
  @classmethod
  def unpack(cls, data:bytes):
    return cls(*(struct.unpack('<fff', data)))

@dataclass(frozen=True)
class Quaternion(Data):
  x: float
  y: float
  z: float
  w: float
  roll: float = field(init=False, default=0.0)
  pitch: float = field(init=False, default=0.0)
  yaw: float = field(init=False, default=0.0)
  def __post_init__(self):
    object.__setattr__(self, "roll", atan2(2*(self.w*self.x + self.y*self.z), 1-2*(self.x**2 + self.y**2)))
    object.__setattr__(self, "pitch", asin(2*(self.w*self.y - self.x*self.z)))
    object.__setattr__(self, "yaw", atan2(2*(self.w*self.z + self.x*self.y), 1-2*(self.y**2 + self.z**2)))
  @classmethod
  def unpack(cls, data:bytes):
    return cls(*(struct.unpack('<ffff', data)))

class MoCapData:
  ...

@dataclass(frozen=True)
class Frame_prefix(MoCapData):
  frame_number: int

@dataclass(slots=True)
class Marker_data:
  name: str
  num_markers: int
  positions: Tuple[Position, ...]

@dataclass(slots=True)
class Marker_set_data(MoCapData):
  num_marker_sets: int
  marker_sets: Tuple[Marker_data, ...]
  marker_sets_d: Dict[str, Marker_data] = field(init=False)

  def __post_init__(self):
    object.__setattr__(self, "marker_sets_d", { instance.name:instance for instance in self.marker_sets})

@dataclass(slots=True)
class Legacy_marker_set_data(MoCapData):
  num_markers: int
  positions: Tuple[Position, ...]

@dataclass(slots=True)
class Rigid_body:
  identifier: int
  pos: Position
  rot: Quaternion
  err: float
  tracking: bool

@dataclass(slots=True)
class Rigid_body_data(MoCapData):
  num_rigid_bodies: int
  rigid_bodies: Tuple[Rigid_body, ...]
  rigid_bodies_d: Dict[int, Rigid_body] = field(init=False)
  
  def __post_init__(self):
    object.__setattr__(self, "rigid_bodies_d", { instance.identifier : instance for instance in self.rigid_bodies})

@dataclass(slots=True)
class Skeleton:
  identifier: int
  num_rigid_bodies: int
  rigid_bodies: Tuple[Rigid_body, ...]
  rigid_bodies_d: Dict[int, Rigid_body] = field(init=False)
  
  def __post_init__(self):
    object.__setattr__(self, "rigid_bodies_d", { instance.identifier : instance for instance in self.rigid_bodies})

@dataclass(slots=True)
class Skeleton_data(MoCapData):
  num_skeletons: int
  skeletons: Tuple[Skeleton, ...]

@dataclass(slots=True)
class Asset_RB:
  identifier: int
  pos: Position
  rot: Quaternion
  err: float
  param: int

@dataclass(slots=True)
class Asset_marker:
  identifier: int
  pos: Position
  size: float
  param: int
  residual: float

@dataclass(slots=True)
class Asset:
  identifier: int
  num_rigid_bodies: int
  rigid_bodies: Tuple[Asset_RB, ...]
  num_markers: int
  markers: Tuple[Asset_marker, ...]
  rigid_bodies_d: Dict[int, Asset_RB] = field(init=False)
  markers_d: Dict[int, Asset_marker] = field(init=False)

  def __post_init__(self):
    object.__setattr__(self, "rigid_bodies_d", { instance.identifier : instance for instance in self.rigid_bodies})
    object.__setattr__(self, "markers_d", { instance.identifier : instance for instance in self.markers})

@dataclass(slots=True)
class Asset_data(MoCapData):
  num_assets: int
  assets: Tuple[Asset, ...]
  assets_d: Dict[int, Asset] = field(init=False)

  def __post_init__(self):
    object.__setattr__(self, "assets_d", { instance.identifier : instance for instance in self.assets})

@dataclass(slots=True)
class Labeled_marker:
  identifier: int
  pos: Position
  size: int
  param: int
  residual: float

@dataclass(slots=True)
class Labeled_marker_data(MoCapData):
  num_markers: int
  markers: Tuple[Labeled_marker, ...]
  markers_d: Dict[int, Labeled_marker] = field(init=False)

  def __post_init__(self):
    object.__setattr__(self, "markers_d", { instance.identifier : instance for instance in self.markers})

@dataclass(slots=True)
class Channel:
  num_frames: int
  frames: Tuple[float, ...]

@dataclass(slots=True)
class Force_plate:
  identifier: int
  num_channels: int
  channels: Tuple[Channel, ...]

@dataclass(slots=True)
class Force_plate_data(MoCapData):
  num_force_plates: int
  force_plates: Tuple[Force_plate, ...]
  force_plates_d: Dict[int, Force_plate] = field(init=False)

  def __post_init__(self):
    object.__setattr__(self, "force_plates_d", { instance.identifier : instance for instance in self.force_plates})

@dataclass(slots=True)
class Device:
  identifier: int
  num_channels: int
  channels: Tuple[Channel, ...]

@dataclass(slots=True)
class Device_data(MoCapData):
  num_devices: int
  devices: Tuple[Device, ...]
  devices_d: Dict[int, Device] = field(init=False)
  
  def __post_init__(self):
    object.__setattr__(self, "devices_d", { instance.identifier : instance for instance in self.devices})

@dataclass(slots=True)
class Frame_suffix:
  time_code: int
  time_code_sub: int
  timestamp: float
  camera_mid_exposure: int
  stamp_data: int
  stamp_transmit: int
  recording: bool
  tracked_models_changed: bool
  precision_timestamp_sec: int | None = None
  precision_timestamp_frac_sec: int | None = None

@dataclass(slots=True)
class MoCap:
  prefix_data: Frame_prefix
  marker_set_data: Marker_set_data
  legacy_marker_set_data: Legacy_marker_set_data
  rigid_body_data: Rigid_body_data
  skeleton_data: Skeleton_data
  labeled_marker_data: Labeled_marker_data
  force_plate_data: Force_plate_data
  device_data: Device_data
  suffix_data: Frame_suffix
  asset_data: Asset_data | None = None

@dataclass(slots=True)
class Marker_set_description:
  name: str
  num_markers:int
  markers_names: Tuple[str, ...]

@dataclass(slots=True)
class RB_marker:
  name:str
  identifier: int
  pos: Position

@dataclass(slots=True)
class Rigid_body_description:
  name: str
  identifier: int
  parent_id: int
  pos: Position
  num_markers:int
  markers: Tuple[RB_marker, ...]
  markers_d: Dict[int, RB_marker] = field(init=False)

  def __post_init__(self):
    object.__setattr__(self, "markers_d", { instance.identifier : instance for instance in self.markers})

@dataclass(slots=True)
class Skeleton_description:
  name: str
  identifier: int
  num_rigid_bodies: int
  rigid_bodies: Tuple[Rigid_body_description, ...]
  rigid_bodies_d: Dict[int, Rigid_body_description] = field(init=False)

  def __post_init__(self):
    object.__setattr__(self, "rigid_bodies_d", { instance.identifier : instance for instance in self.rigid_bodies})

@dataclass(slots=True)
class Force_plate_description:
  identifier:int
  serial_number: str
  dimensions: Tuple[float, float]
  origin: Position
  calibration_matrix: Tuple[float, ...]
  corners: Tuple[float, ...]
  plate_type: int
  channel_data_type: int
  num_channels: int
  channels: Tuple[str, ...]

@dataclass(slots=True)
class Device_description:
  identifier: int
  name: str
  serial_number: str
  type: int
  channel_type: int
  num_channels: int
  channels: Tuple[str, ...]

@dataclass(slots=True)
class Camera_description:
  name: str
  pos: Position
  orientation: Quaternion

@dataclass(slots=True)
class Marker_description:
  name: str
  identifier: int
  pos: Position
  size: float
  param: int

@dataclass(slots=True)
class Asset_description:
  name: str
  type: int
  identifier: int
  num_rigid_bodies: int
  rigid_bodies: Tuple[Rigid_body_description, ...]
  num_markers: int
  markers: Tuple[Marker_description, ...]
  rigid_bodies_d: Dict[int, Rigid_body_description] = field(init=False)

  def __post_init__(self):
    object.__setattr__(self, "rigid_bodies_d", { instance.identifier : instance for instance in self.rigid_bodies})

@dataclass(slots=True)
class Descriptors:
  """
    Object for storing descriptions
  """  
  marker_set_description: Dict[str, Marker_set_description] = field(init=False, default_factory=dict)
  rigid_body_description: Dict[int, Rigid_body_description] = field(init=False, default_factory=dict)
  skeleton_description: Dict[int, Skeleton_description] = field(init=False, default_factory=dict)
  force_plate_description: Dict[str, Force_plate_description] = field(init=False, default_factory=dict)
  device_description: Dict[str, Device_description] = field(init=False, default_factory=dict)
  camera_description: Dict[str, Camera_description] = field(init=False, default_factory=dict)
  asset_description: Dict[int, Asset_description] = field(init=False, default_factory=dict)


import new_natnet_client.NatNetTypes as NatNetTypes
from typing import Callable, Iterable, Tuple, Dict
from collections import deque
from struct import unpack
from itertools import batched

class DataUnpackerV3_0:
  rigid_body_lenght: int = 38
  marker_lenght: int = 26
  frame_suffix_lenght: int = 42
  @classmethod
  def unpack_data_size(cls, data: bytes) -> Tuple[int, int]:
    return 0, 0

  @classmethod
  def unpack_frame_prefix_data(cls, data: bytes) -> Tuple[NatNetTypes.Frame_prefix, int]:
    offset = 0
    prefix = NatNetTypes.Frame_prefix(int.from_bytes(data[offset : (offset:=offset+4)], byteorder='little', signed=True))
    return prefix, offset

  @classmethod
  def unpack_marker_set_data(cls, data: bytes) -> Tuple[NatNetTypes.Marker_set_data, int]:
    offset = 0
    num_marker_sets = int.from_bytes(data[offset : (offset:=offset+4)], byteorder='little', signed=True)
    _, tmp_offset = cls.unpack_data_size(data)
    offset += tmp_offset
    markers: Iterable[NatNetTypes.Marker_data] = deque()
    position_unpacker: Callable[[Iterable[int]], NatNetTypes.Position] = lambda position_data: NatNetTypes.Position.unpack(bytes(position_data))
    for _ in range(num_marker_sets):
      name, _, _ = data[offset:].partition(b'\0')
      offset += len(name) + 1
      name = str(name, encoding="utf-8")
      num_markers = int.from_bytes(data[offset : (offset:=offset+4)], byteorder='little', signed=True)
      positions = tuple(map(
        position_unpacker,
        batched(data[offset : (offset:=offset+(12*num_markers))], 12)
      ))
      markers.append(NatNetTypes.Marker_data(name, num_markers, positions))
    marker_sets = tuple(markers)
    return NatNetTypes.Marker_set_data(num_marker_sets, marker_sets), offset

  @classmethod
  def unpack_legacy_other_markers(cls, data: bytes) -> Tuple[NatNetTypes.Legacy_marker_set_data,int]:
    offset = 0
    num_markers = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    _, tmp_offset = cls.unpack_data_size(data)
    offset += tmp_offset
    positions = deque(map(
      lambda position_data: NatNetTypes.Position.unpack(bytes(position_data)),
      batched(data[offset:(offset:=offset+(12*num_markers))], 12)
    ))
    positions = tuple(positions)
    return NatNetTypes.Legacy_marker_set_data(num_markers, positions), offset

  @classmethod
  def unpack_rigid_body(cls, data: bytes) -> NatNetTypes.Rigid_body:
    offset = 0
    identifier = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    pos = NatNetTypes.Position.unpack(data[offset:(offset:=offset+12)])
    rot = NatNetTypes.Quaternion.unpack(data[offset:(offset:=offset+16)])
    err = unpack('<f', data[offset:(offset:=offset+4)])[0]
    param: int = unpack( 'h', data[offset:(offset:=offset+2)])[0]
    tracking = bool(param & 0x01)
    return NatNetTypes.Rigid_body(identifier, pos, rot, err, tracking)

  @classmethod
  def unpack_rigid_body_data(cls, data: bytes) -> Tuple[NatNetTypes.Rigid_body_data, int]:
    offset = 0
    num_rigid_bodies = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    _, tmp_offset = cls.unpack_data_size(data)
    offset += tmp_offset
    rigid_bodies = tuple(map(
        lambda rigid_body_data: cls.unpack_rigid_body(bytes(rigid_body_data)),
        batched(data[offset:(offset:=offset+(cls.rigid_body_lenght*num_rigid_bodies))], cls.rigid_body_lenght) 
    ))
    return NatNetTypes.Rigid_body_data(num_rigid_bodies, rigid_bodies), offset
  
  @classmethod
  def unpack_skeleton(cls, data: bytes) -> Tuple[NatNetTypes.Skeleton, int]:
    offset = 0
    identifier = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    num_rigid_bodies = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    rigid_bodies = tuple(map(
        lambda rigid_body_data: cls.unpack_rigid_body(bytes(rigid_body_data)),
        batched(data[offset:(offset:=offset+(cls.rigid_body_lenght*num_rigid_bodies))], cls.rigid_body_lenght) 
    ))
    return NatNetTypes.Skeleton(identifier, num_rigid_bodies, rigid_bodies), offset

  @classmethod
  def unpack_skeleton_data(cls, data: bytes) -> Tuple[NatNetTypes.Skeleton_data, int]:
    offset = 0
    num_skeletons = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    _, tmp_offset = cls.unpack_data_size(data)
    offset += tmp_offset
    skeletons: Iterable[NatNetTypes.Skeleton] = deque()
    for _ in range(num_skeletons):
      skeleton, tmp_offset = cls.unpack_skeleton(data[offset:])
      offset += tmp_offset
      skeletons.append(skeleton)
    skeletons = tuple(skeletons)
    return NatNetTypes.Skeleton_data(num_skeletons, skeletons), offset

  @classmethod
  def unpack_asset_rigid_body(cls, data: bytes) -> NatNetTypes.Asset_RB:
    raise NotImplementedError("Subclasses must implement the unpack method")

  @classmethod
  def unpack_asset_marker(cls, data: bytes) -> NatNetTypes.Asset_marker:
    raise NotImplementedError("Subclasses must implement the unpack method")

  @classmethod
  def unpack_asset(cls, data: bytes) -> Tuple[NatNetTypes.Asset, int]:
    raise NotImplementedError("Subclasses must implement the unpack method")

  @classmethod
  def unpack_asset_data(cls, data: bytes) -> Tuple[NatNetTypes.Asset_data, int]:
    raise NotImplementedError("Subclasses must implement the unpack method")

  @classmethod
  def decode_marker_id(cls, identifier: int) -> Tuple[int, int]:
    return (
      identifier >> 16,
      identifier & 0x0000ffff
    )

  @classmethod
  def unpack_labeled_marker(cls, data: bytes) -> NatNetTypes.Labeled_marker:
    offset = 0
    identifier = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    pos = NatNetTypes.Position.unpack(data[offset:(offset:=offset+12)])
    size = unpack('<f', data[offset:(offset:=offset+4)])[0]
    param = unpack( 'h', data[offset:(offset:=offset+2)])[0]
    residual = unpack('<f', data[offset:(offset:=offset+4)])[0] * 1000.0
    return NatNetTypes.Labeled_marker(identifier, pos, size, param, residual)

  @classmethod
  def unpack_labeled_marker_data(cls, data: bytes) -> Tuple[NatNetTypes.Labeled_marker_data, int]:
    offset = 0
    num_markers = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    _, tmp_offset = cls.unpack_data_size(data)
    offset += tmp_offset
    markers = tuple(map(
      lambda marker_data: cls.unpack_labeled_marker(bytes(marker_data)),
      batched(data[offset:(offset:=offset+(cls.marker_lenght*num_markers))],cls.marker_lenght)
    ))
    return NatNetTypes.Labeled_marker_data(num_markers, markers), offset

  @classmethod
  def unpack_channels(cls, data: bytes, num_channels: int) -> Tuple[Tuple[NatNetTypes.Channel, ...], int]:
    offset = 0
    channels: Iterable[NatNetTypes.Channel] = deque()
    frame_unpacker: Callable[[Iterable[int]], float] = lambda frame_data: unpack('<f', bytes(frame_data))[0]
    for _ in range(num_channels):
      num_frames = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
      frames = tuple(map(
        frame_unpacker,
        batched(data[offset:(offset:=offset+(4*num_frames))],4)
      ))
      channels.append(NatNetTypes.Channel(num_frames, frames))
    return tuple(channels), offset

  @classmethod
  def unpack_force_plate_data(cls, data: bytes) -> Tuple[NatNetTypes.Force_plate_data, int]:
    offset = 0
    num_force_plates = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    _, tmp_offset = cls.unpack_data_size(data)
    offset += tmp_offset
    force_plates: Iterable[NatNetTypes.Force_plate] = deque()
    for _ in range(num_force_plates):
      identifier = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
      num_channels = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
      channels, tmp_offset = cls.unpack_channels(data[offset:], num_channels)
      offset += tmp_offset
      force_plates.append(NatNetTypes.Force_plate(identifier, num_channels, channels))
    force_plates = tuple(force_plates)
    return NatNetTypes.Force_plate_data(num_force_plates, force_plates), offset

  @classmethod
  def unpack_device_data(cls, data: bytes) -> Tuple[NatNetTypes.Device_data, int]:
    offset = 0
    num_devices = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    _, tmp_offset = cls.unpack_data_size(data)
    offset += tmp_offset
    devices: Iterable[NatNetTypes.Device] = deque()
    for _ in range(num_devices):
      identifier = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
      num_channels = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
      channels, tmp_offset = cls.unpack_channels(data[offset:], num_channels)
      offset += tmp_offset
      devices.append(NatNetTypes.Device(identifier, num_channels, channels))
    devices = tuple(devices)
    return NatNetTypes.Device_data(num_devices, devices), offset

  @classmethod
  def unpack_frame_suffix_data(cls, data: bytes) -> NatNetTypes.Frame_suffix:
    offset = 0
    time_code = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    time_code_sub = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    timestamp = unpack('<d', data[offset:(offset:=offset+8)])[0]
    camera_mid_exposure = int.from_bytes(data[offset:(offset:=offset+8)], byteorder='little', signed=True)
    stamp_data = int.from_bytes(data[offset:(offset:=offset+8)], byteorder='little', signed=True)
    stamp_transmit = int.from_bytes(data[offset:(offset:=offset+8)], byteorder='little', signed=True)
    param = unpack( 'h', data[offset:(offset:=offset+2)])[0]
    recording = bool(param & 0x01)
    tracked_models_changed = bool(param & 0x02)
    return NatNetTypes.Frame_suffix(time_code, time_code_sub, timestamp, camera_mid_exposure, stamp_data, stamp_transmit, recording, tracked_models_changed)

  @classmethod
  def unpack_mocap_data(cls, data: bytes) -> NatNetTypes.MoCap:
    offset = 0
    tmp_offset = 0

    prefix_data, tmp_offset = cls.unpack_frame_prefix_data(data[offset:])
    offset += tmp_offset

    marker_set_data, tmp_offset = cls.unpack_marker_set_data(data[offset:])
    offset += tmp_offset

    legacy_marker_set_data, tmp_offset = cls.unpack_legacy_other_markers(data[offset:])
    offset += tmp_offset

    rigid_body_data, tmp_offset = cls.unpack_rigid_body_data(data[offset:])
    offset += tmp_offset

    skeleton_data, tmp_offset = cls.unpack_skeleton_data(data[offset:])
    offset += tmp_offset

    labeled_marker_data, tmp_offset = cls.unpack_labeled_marker_data(data[offset:])
    offset += tmp_offset

    force_plate_data, tmp_offset = cls.unpack_force_plate_data(data[offset:])
    offset += tmp_offset

    device_data, tmp_offset = cls.unpack_device_data(data[offset:])
    offset += tmp_offset

    suffix_data = cls.unpack_frame_suffix_data(data[offset:])

    return NatNetTypes.MoCap(prefix_data, marker_set_data, legacy_marker_set_data, rigid_body_data, skeleton_data, labeled_marker_data, force_plate_data, device_data, suffix_data)

  @classmethod
  def unpack_marker_set_description(cls, data: bytes) -> Tuple[Dict[str, NatNetTypes.Marker_set_description], int]:
    offset = 0
    name, _, _ = data[offset:].partition(b'\0')
    offset += len(name) + 1
    name = str(name, encoding="utf-8")
    num_markers = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    markers_names: Iterable[str] = deque()
    for _ in range(num_markers):
      marker_name, _, _ = data[offset:].partition(b'\0')
      offset += len(name) + 1
      markers_names.append(str(marker_name, encoding="utf-8"))
    markers_names = tuple(markers_names)
    return {name : NatNetTypes.Marker_set_description(name, num_markers, markers_names)}, offset

  @classmethod
  def unpack_rigid_body_description(cls, data: bytes) -> Tuple[Dict[int, NatNetTypes.Rigid_body_description], int]:
    offset = 0
    name, _, _ = data[offset:].partition(b'\0')
    offset += len(name) + 1
    name = str(name, encoding="utf-8")
    identifier = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    parent_id = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    pos = NatNetTypes.Position.unpack(data[offset:(offset:=offset+12)])
    num_markers = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    offset_pos = offset
    offset_id = offset_pos + (12*num_markers)
    offset_name = offset_id + (4*num_markers)
    marker_name = ""
    markers: Iterable[NatNetTypes.RB_marker] = deque()
    for _ in range(num_markers):
      marker_pos = NatNetTypes.Position.unpack(data[offset_pos:(offset_pos:=offset_pos+12)])
      marker_id = int.from_bytes(data[offset_id:(offset_id:=offset_id+4)], byteorder='little', signed=True)
      markers.append(NatNetTypes.RB_marker(marker_name, marker_id, marker_pos))
    markers = tuple(markers)
    return {identifier : NatNetTypes.Rigid_body_description(name, identifier, parent_id, pos, num_markers, markers)}, offset_name

  @classmethod
  def unpack_skeleton_description(cls, data: bytes) -> Tuple[Dict[int, NatNetTypes.Skeleton_description], int]:
    offset = 0
    name, _, _ = data[offset:].partition(b'\0')
    offset += len(name) + 1
    name = str(name, encoding="utf-8")
    identifier = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    num_rigid_bodies = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    rigid_bodies: Iterable[NatNetTypes.Rigid_body_description] = deque()
    for _ in range(num_rigid_bodies):
      d, offset_tmp = cls.unpack_rigid_body_description(data[offset:])
      rigid_body = list(d.values())[0]
      rigid_bodies.append(rigid_body)
      offset += offset_tmp
    rigid_bodies = tuple(rigid_bodies)
    return {identifier : NatNetTypes.Skeleton_description(name, identifier, num_rigid_bodies, rigid_bodies)}, offset

  @classmethod
  def unpack_force_plate_description(cls, data: bytes) -> Tuple[Dict[str, NatNetTypes.Force_plate_description], int]:
    offset = 0
    identifier = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little',  signed=True)

    serial_number, _, _ = data[offset:].partition(b'\0')
    offset += len(serial_number) + 1
    serial_number = str(serial_number, encoding='utf-8')

    f_width: float = unpack('<f', data[offset:(offset:=offset+4)])[0]
    f_length: float = unpack('<f', data[offset:(offset:=offset+4)])[0]
    dimensions = (f_width, f_length)

    origin = NatNetTypes.Position.unpack(data[offset:(offset:=offset+12)])

    # Not tested
    calibration_matrix = tuple(unpack('<f', data[offset:(offset:=offset+4)])[0] for _ in range(12*12))
    corners = tuple(unpack('<f', data[offset:(offset:=offset+4)])[0] for _ in range(4*3))

    plate_type = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little',  signed=True)
    channel_data_type = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little',  signed=True)
    num_channels = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little',  signed=True)
    
    channels: Iterable[str] = deque()
    for _ in range(num_channels):
      channel_name, _, _ = data[offset:].partition(b'\0')
      offset += len(channel_name) + 1
      channels.append(str(channel_name, encoding='utf-8'))
    channels = tuple(channels)
    return {serial_number : NatNetTypes.Force_plate_description(identifier, serial_number, dimensions, origin, calibration_matrix, corners, plate_type, channel_data_type, num_channels, channels)}, offset

  @classmethod
  def unpack_device_description(cls, data: bytes) -> Tuple[Dict[str, NatNetTypes.Device_description], int]:
    offset = 0
    
    identifier = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little',  signed=True)
    
    name, _, _ = data[offset:].partition(b'\0')
    offset += len(name) + 1
    name = str(name, encoding='utf-8')
    
    serial_number, _, _ = data[offset:].partition(b'\0')
    offset += len(serial_number) + 1
    serial_number = str(serial_number, encoding='utf-8')

    device_type = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little',  signed=True)
    channel_data_type = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little',  signed=True)
    num_channels = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little',  signed=True)
    channels: Iterable[str] = deque()
    for _ in range(num_channels):
      channel_name,_,_ = data[offset:].partition(b'\0')
      offset += len(channel_name) + 1
      channels.append(str(channel_name, encoding='utf-8'))
    channels = tuple(channels)
    return {serial_number : NatNetTypes.Device_description(identifier, name, serial_number, device_type, channel_data_type, num_channels, channels)}, offset

  @classmethod
  def unpack_camera_description(cls, data: bytes) -> Tuple[Dict[str, NatNetTypes.Camera_description], int]:
    offset = 0
    name, _, _ = data[offset:].partition(b'\0')
    offset += len(name) + 1
    name = str(name, encoding="utf-8")
    pos = NatNetTypes.Position.unpack(data[offset:(offset:=offset+12)])
    orientation = NatNetTypes.Quaternion.unpack(data[offset:(offset:=offset+16)])
    return {name : NatNetTypes.Camera_description(name, pos, orientation)}, offset

  @classmethod
  def unpack_marker_description(cls, data: bytes) -> Tuple[Dict[int, NatNetTypes.Marker_description], int]:
    offset = 0
    name, _, _ = data[offset:].partition(b'\0')
    offset += len(name) + 1
    name = str(name, encoding="utf-8")
    identifier = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little',  signed=True)
    pos = NatNetTypes.Position.unpack(data[offset:(offset:=offset+12)])
    size = unpack('<f', data[offset:(offset:=offset+4)])[0]
    param = unpack( 'h', data[offset:(offset:=offset+2)])[0]
    return {identifier : NatNetTypes.Marker_description(name, identifier, pos, size, param)}, offset

  @classmethod
  def unpack_asset_description(cls, data: bytes) -> Tuple[Dict[int, NatNetTypes.Asset_description], int]:
    offset = 0
    name, _, _ = data[offset:].partition(b'\0')
    offset += len(name) + 1
    name = str(name, encoding="utf-8")
    asset_type = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little',  signed=True)
    identifier = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little',  signed=True)
    num_rigid_bodies = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little',  signed=True)
    rigid_bodies: Iterable[NatNetTypes.Rigid_body_description] = deque()
    for _ in range(num_rigid_bodies):
      d, offset_tmp = cls.unpack_rigid_body_description(data[offset:])
      rigid_body = list(d.values())[0]
      rigid_bodies.append(rigid_body)
      offset += offset_tmp
    rigid_bodies = tuple(rigid_bodies)
    num_markers = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little',  signed=True)
    markers: Iterable[NatNetTypes.Marker_description] = deque()
    for _ in range(num_markers):
      d,offset_tmp = cls.unpack_marker_description(data[offset:])
      marker = list(d.values())[0]
      markers.append(marker)
      offset += offset_tmp
    markers = tuple(markers)
    return {identifier : NatNetTypes.Asset_description(name, asset_type, identifier, num_rigid_bodies, rigid_bodies, num_markers, markers)}, offset

class DataUnpackerV4_1(DataUnpackerV3_0):
  asset_rigid_body_lenght: int = 38
  asset_marker_lenght: int = 26
  frame_suffix_lenght: int = 50

  @classmethod
  def unpack_data_size(cls, data: bytes) -> Tuple[int, int]:
    offset = 0
    size_in_bytes = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    return size_in_bytes, offset

  @classmethod
  def unpack_asset_rigid_body(cls, data: bytes) -> NatNetTypes.Asset_RB:
    offset = 0
    identifier = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    pos = NatNetTypes.Position.unpack(data[offset:(offset:=offset+12)])
    rot = NatNetTypes.Quaternion.unpack(data[offset:(offset:=offset+16)])
    err = unpack('<f', data[offset:(offset:=offset+4)])[0]
    param = unpack( 'h', data[offset:(offset:=offset+2)])[0]
    return NatNetTypes.Asset_RB(identifier, pos, rot, err, param)

  @classmethod
  def unpack_asset_marker(cls, data: bytes) -> NatNetTypes.Asset_marker:
    offset = 0
    identifier = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    pos = NatNetTypes.Position.unpack(data[offset:(offset:=offset+12)])
    size = unpack('<f', data[offset:(offset:=offset+4)])[0]
    param = unpack( 'h', data[offset:(offset:=offset+2)])[0]
    residual = unpack('<f', data[offset:(offset:=offset+4)])[0]
    return NatNetTypes.Asset_marker(identifier, pos, size, param, residual)

  @classmethod
  def unpack_asset(cls, data: bytes) -> Tuple[NatNetTypes.Asset, int]:
    offset = 0
    identifier = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    num_rigid_bodies = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    rigid_bodies = tuple(map(
      lambda rigid_body_data: cls.unpack_asset_rigid_body(bytes(rigid_body_data)),
      batched(data[offset:(offset:=offset+(cls.asset_rigid_body_lenght*num_rigid_bodies))], cls.asset_rigid_body_lenght)
    ))
    num_markers = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    markers = tuple(map(
      lambda marker_data: cls.unpack_asset_marker(bytes(marker_data)),
      batched(data[offset:(offset:=offset+(cls.asset_marker_lenght*num_markers))], cls.asset_marker_lenght)
    ))
    return NatNetTypes.Asset(identifier, num_rigid_bodies, rigid_bodies, num_markers, markers), offset

  @classmethod
  def unpack_asset_data(cls, data: bytes) -> Tuple[NatNetTypes.Asset_data, int]:
    offset = 0
    num_assets = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    _, tmp_offset = cls.unpack_data_size(data)
    offset += tmp_offset
    assets: Iterable[NatNetTypes.Asset] = deque()
    for _ in range(num_assets):
      asset, tmp_offset = cls.unpack_asset(data[offset:])
      offset += tmp_offset
      assets.append(asset)
    assets = tuple(assets)
    return NatNetTypes.Asset_data(num_assets, assets), offset

  @classmethod
  def unpack_frame_suffix_data(cls, data: bytes) -> NatNetTypes.Frame_suffix:
    offset = 0
    time_code = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    time_code_sub = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    timestamp = unpack('<d', data[offset:(offset:=offset+8)])[0]
    camera_mid_exposure = int.from_bytes(data[offset:(offset:=offset+8)], byteorder='little', signed=True)
    stamp_data = int.from_bytes(data[offset:(offset:=offset+8)], byteorder='little', signed=True)
    stamp_transmit = int.from_bytes(data[offset:(offset:=offset+8)], byteorder='little', signed=True)
    precision_timestamp_sec = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    precision_timestamp_frac_sec = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    param = unpack( 'h', data[offset:(offset:=offset+2)])[0]
    recording = bool(param & 0x01)
    tracked_models_changed = bool(param & 0x02)
    return NatNetTypes.Frame_suffix(time_code, time_code_sub, timestamp, camera_mid_exposure, stamp_data, stamp_transmit, recording, tracked_models_changed, precision_timestamp_sec, precision_timestamp_frac_sec)

  @classmethod
  def unpack_mocap_data(cls, data: bytes) -> NatNetTypes.MoCap:
    offset = 0
    tmp_offset = 0

    prefix_data, tmp_offset = cls.unpack_frame_prefix_data(data[offset:])
    offset += tmp_offset

    marker_set_data, tmp_offset = cls.unpack_marker_set_data(data[offset:])
    offset += tmp_offset

    legacy_marker_set_data, tmp_offset = cls.unpack_legacy_other_markers(data[offset:])
    offset += tmp_offset

    rigid_body_data, tmp_offset = cls.unpack_rigid_body_data(data[offset:])
    offset += tmp_offset

    skeleton_data, tmp_offset = cls.unpack_skeleton_data(data[offset:])
    offset += tmp_offset

    asset_data, tmp_offset = cls.unpack_asset_data(data[offset:])
    offset += tmp_offset

    labeled_marker_data, tmp_offset = cls.unpack_labeled_marker_data(data[offset:])
    offset += tmp_offset

    force_plate_data, tmp_offset = cls.unpack_force_plate_data(data[offset:])
    offset += tmp_offset

    device_data, tmp_offset = cls.unpack_device_data(data[offset:])
    offset += tmp_offset

    suffix_data = cls.unpack_frame_suffix_data(data[offset:])

    return NatNetTypes.MoCap(prefix_data, marker_set_data, legacy_marker_set_data, rigid_body_data, skeleton_data, labeled_marker_data, force_plate_data, device_data, suffix_data, asset_data)

  @classmethod
  def unpack_rigid_body_description(cls, data: bytes) -> Tuple[Dict[int, NatNetTypes.Rigid_body_description], int]:
    d, offset = super().unpack_rigid_body_description(data)
    rb_desc = tuple(d.values())[0]
    for marker in rb_desc.markers:
      name, _, _ = data[offset:].partition(b'\0')
      offset += len(name) + 1
      marker.name = str(name, encoding="utf-8")
    return d, offset


from dataclasses import dataclass, field, FrozenInstanceError, InitVar, asdict
from collections import deque
import struct
from types import NoneType, TracebackType
from typing import Any, Optional, Tuple, Dict, Callable, Iterator, Type
import socket
import logging
from threading import Thread, Lock
from copy import copy, deepcopy
import time

from new_natnet_client.NatNetTypes import NAT_Messages, NAT_Data, MoCap, Descriptors
from new_natnet_client.Unpackers import DataUnpackerV3_0, DataUnpackerV4_1

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(levelname)s - %(message)s')

@dataclass(slots=True)
class Server_info:
  application_name: str
  version: Tuple[int, ...]
  nat_net_major: int
  nat_net_minor: int

@dataclass(kw_only=True)
class NatNetClient:
  server_address: str = "127.0.0.1"
  local_ip_address: str = "127.0.0.1"
  use_multicast: bool = True
  multicast_address: str ="239.255.42.99"
  command_port: int = 1510
  data_port: int = 1511
  max_buffer_size: InitVar[int] = 255
  __mocap_bytes: bytes | None = field(init=False, default=None)
  __last_new_data_time: int = field(init=False, default=0)
  __new_data_flag: bool = field(init=False, default=False)
  __descriptors: Descriptors = field(init=False, default_factory=Descriptors)
  __can_change_bitstream: bool = field(init=False, default=False)
  __running: bool = field(init=False, default=False)
  __command_socket: socket.socket | None = field(init=False, repr=False, default=None)
  __data_socket: socket.socket | None = field(init=False, repr=False, default=None)
  __freeze: bool = field(init=False, repr=False, default=False)

  # TODO: Add bitstream change support

  @property
  def connected(self) -> bool:
    return self.__running

  @property
  def server_info(self) -> Server_info:
    with self.__server_info_lock:
      return copy(self.__server_info)

  @property
  def last_mocap_data(self) -> None | MoCap:
    if  self.__mocap_bytes is None:
      return None
    return self.__unpacker.unpack_mocap_data(self.__mocap_bytes)

  @property
  def MoCap(self) -> Iterator[MoCap]:
    if  self.__mocap_bytes is None: raise StopIteration
    while True:
      with self.__mocap_bytes_lock:
        if self.__new_data_flag:
          yield self.__unpacker.unpack_mocap_data(self.__mocap_bytes)
          self.__new_data_flag = False

  @property
  def last_new_data_time(self):
    return self.__last_new_data_time

  @property
  def server_responses(self) -> deque[int | str]:
    with self.__server_responses_lock:
      return self.__server_responses.copy()

  @property
  def server_messages(self) -> deque[str]:
    with self.__server_messages_lock:
      return self.__server_messages.copy()

  @property
  def buffer_size(self) -> int:
    """
      Buffer size for messages/responses queues
    """
    return self.__max_buffer_size

  @buffer_size.setter
  def buffer_size(self, max_len: int) -> None:
    self.__max_buffer_size = max_len
    with self.__server_messages_lock:
      self.__server_messages = deque(self.__server_messages,maxlen=max_len)
    with self.__server_responses_lock:
      self.__server_responses = deque(self.__server_responses,maxlen=max_len)

  @property
  def frozen_descriptors(self) -> Descriptors:
    return deepcopy(self.__descriptors)
  
  @property
  def descriptors(self) -> Descriptors:
    return self.__descriptors

  @staticmethod
  def create_socket(ip: str, proto: int, port: int = 0) -> socket.socket | None:
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, proto)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    try:
      # Connect to the IP with a dynamically assigned port
      sock.bind((ip, port))
      sock.settimeout(3)
      return sock
    except socket.error as msg:
      logging.error(msg)
      sock.close()

  def __get_message_id(self, data: bytes) -> int:
    message_id = int.from_bytes( data[0:2], byteorder='little',  signed=True )
    return message_id

  def __create__command_socket(self) -> bool:
    ip = self.local_ip_address
    proto = socket.IPPROTO_UDP
    if self.use_multicast:
      ip = ''
      # Let system decide protocol
      proto = 0
    self.__command_socket = self.create_socket(ip, proto)
    if type(self.__command_socket) == NoneType:
      logging.info(f"Command socket. Check Motive/Server mode requested mode agreement.  {self.use_multicast = } ")
      return False
    if self.use_multicast:
      # set to broadcast mode
      self.__command_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    return True

  def __create__data_socket(self) -> bool:
    ip = ''
    proto = socket.IPPROTO_UDP
    port = 0
    if self.use_multicast:
      ip = self.local_ip_address
      proto = 0
      port = self.data_port
    self.__data_socket = self.create_socket(ip, proto, port)
    if type(self.__data_socket) == NoneType:
      logging.info(f"Data socket. Check Motive/Server mode requested mode agreement.  {self.use_multicast = } ")
      return False
    if self.use_multicast or self.multicast_address != "255.255.255.255":
      self.__data_socket.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, socket.inet_aton(self.multicast_address) + socket.inet_aton(self.local_ip_address))
    return True

  def __post_init__(self, max_buffer_size: int) -> None:
    self.__running_lock = Lock()
    self.__command_socket_lock = Lock()
    self.__server_info_lock = Lock()
    self.__server_responses_lock = Lock()
    self.__server_messages_lock = Lock()
    self.__mocap_bytes_lock = Lock()

    self.__max_buffer_size = max_buffer_size
    # Buffer for server responses
    self.__server_responses: deque[int | str] = deque(maxlen=self.__max_buffer_size)
    # Buffer for server messages
    self.__server_messages: deque[str] = deque(maxlen=self.__max_buffer_size)

    self.__server_info: Server_info = Server_info("None", (0,0,0,0), 0,0)
    
    # Map unpacking methods with respective messages
    self.__mapped: Dict[NAT_Messages, Callable[[bytes,int], int] ] = {
      NAT_Messages.FRAME_OF_DATA: self.__unpack_mocap_data,
      NAT_Messages.MODEL_DEF: self.__unpack_data_descriptions,
      NAT_Messages.SERVER_INFO: self.__unpack_server_info,
      NAT_Messages.RESPONSE: self.__unpack_server_response,
      NAT_Messages.MESSAGE_STRING: self.__unpack_server_message,
      NAT_Messages.UNRECOGNIZED_REQUEST: self.__unpack_unrecognized_request,
      NAT_Messages.UNDEFINED: self.__unpack_undefined_nat_message
    }

    self.__update_unpacker_version()

  def __setattr__(self, name: str, value: Any) -> None:
    if self.__freeze and name in (
      "server_address",
      "local_ip_address",
      "use_multicast",
      "multicast_address",
      "command_port",
      "data_port",
    ):
      raise FrozenInstanceError("This attribute can't be changed because client is already connected")
    super().__setattr__(name, value)

  def connect(self) -> None:
    if self.__running or not self.__create__command_socket() or not self.__create__data_socket(): return
    logging.info("Client connected")
    self.__running = True
    self.__data_thread = Thread(target=self.__data_thread_function)
    self.__data_thread.start()
    self.__command_thread = Thread(target=self.__command_thread_function)
    self.__command_thread.start()
    time.sleep(1) # wait to get threads running for receiving data
    self.send_request(NAT_Messages.CONNECT,"")

  def __enter__(self):
    self.connect()
    return self

  def __exit__(self, exc_type: Optional[Type[BaseException]], exc_value: Optional[BaseException], traceback: Optional[TracebackType]) -> None:
    if self.__running:
      self.shutdown()

  def send_request(self, NAT_command: NAT_Messages, command: str) -> int:
    """
      Send request to server
      
      Returns:
      ---
        int: number of bytes send, (-1) if something went wrong
    """
    if type(self.__command_socket) == NoneType or NAT_command == NAT_Messages.UNDEFINED: return -1
    packet_size: int = 0
    if  NAT_command == NAT_Messages.KEEP_ALIVE or \
        NAT_command == NAT_Messages.REQUEST_MODEL_DEF or \
        NAT_command == NAT_Messages.REQUEST_FRAME_OF_DATA:
      command = ""
    elif NAT_command == NAT_Messages.REQUEST:
      packet_size = len(command) + 1
    elif NAT_command == NAT_Messages.CONNECT:
      tmp_version = [4,1,0,0]
      command = ("Ping".ljust(265, '\x00') + \
                  chr(tmp_version[0]) + \
                  chr(tmp_version[1]) + \
                  chr(tmp_version[2]) + \
                  chr(tmp_version[3]) + \
                  '\x00')
      packet_size = len(command) + 1
    data = NAT_command.value.to_bytes(2, byteorder="little", signed=True)
    data += packet_size.to_bytes(2, byteorder='little',  signed=True)
    data += command.encode("utf-8")
    data += b'\0'
    with self.__command_socket_lock:
      return self.__command_socket.sendto(data, (self.server_address, self.command_port))

  def send_command(self, command: str) -> bool:
    """
      Tries to send the command 3 times

      Returns:
        bool: whether the was send successfully
    """
    res:int = -1
    for _ in range(3):
      res = self.send_request(NAT_Messages.REQUEST, command)
      if res != -1:
        break
    return res != -1

  def __update_unpacker_version(self) -> None:
    """
      Changes unpacker version based on server's bit stream version
    """
    self.__unpacker = DataUnpackerV3_0
    with self.__server_info_lock:
      if (self.__server_info.nat_net_major == 4 and self.__server_info.nat_net_minor >= 1) or self.__server_info.nat_net_major == 0:
        self.__unpacker = DataUnpackerV4_1

  def __unpack_mocap_data(self, data: bytes, packet_size: int) -> int:
    self.__last_new_data_time = time.time_ns()
    with self.__mocap_bytes_lock:
      self.__new_data_flag = True
      self.__mocap_bytes = data
    return packet_size

  def __unpack_data_descriptions(self, data: bytes, packet_size: int) -> int:
    offset = 0
    dataset_count = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
    size_in_bytes = -1
    for _ in range(dataset_count):
      tag = int.from_bytes(data[offset:(offset:=offset+4)], byteorder='little', signed=True)
      data_description_type = NAT_Data(tag)
      if self.__unpacker == DataUnpackerV4_1:
        size_in_bytes = int.from_bytes( data[offset:(offset:=offset+4)], byteorder='little',  signed=True )
      match data_description_type:
        case NAT_Data.MARKER_SET:
          description, tmp_offset = self.__unpacker.unpack_marker_set_description(data[offset:])
          self.__descriptors.marker_set_description.update(description)
        case NAT_Data.RIGID_BODY:
          description, tmp_offset = self.__unpacker.unpack_rigid_body_description(data[offset:])
          self.__descriptors.rigid_body_description.update(description)
        case NAT_Data.SKELETON:
          description, tmp_offset = self.__unpacker.unpack_skeleton_description(data[offset:])
          self.__descriptors.skeleton_description.update(description)
        case NAT_Data.FORCE_PLATE:
          description, tmp_offset = self.__unpacker.unpack_force_plate_description(data[offset:])
          self.__descriptors.force_plate_description.update(description)
        case NAT_Data.DEVICE:
          description, tmp_offset = self.__unpacker.unpack_device_description(data[offset:])
          self.__descriptors.device_description.update(description)
        case NAT_Data.CAMERA:
          description, tmp_offset = self.__unpacker.unpack_camera_description(data[offset:])
          self.__descriptors.camera_description.update(description)
        case NAT_Data.ASSET:
          description, tmp_offset = self.__unpacker.unpack_asset_description(data[offset:])
          self.__descriptors.asset_description.update(description)
        case NAT_Data.UNDEFINED:
          logging.error(f"ID: {tag} - Size: {size_in_bytes}")
          continue
      offset += tmp_offset
    return offset

  def __unpack_server_info(self, data: bytes, __: int) -> int:
    offset = 0
    application_name, _, _ = data[offset:(offset:=offset+256)].partition(b'\0')
    application_name = str(application_name, "utf-8")
    version = struct.unpack( 'BBBB', data[offset:(offset:=offset+4)] )
    nat_net_major, nat_net_minor, _, _ = struct.unpack( 'BBBB', data[offset:(offset:=offset+4)] )
    with self.__server_info_lock:
      self.__server_info = Server_info(application_name, version, nat_net_major, nat_net_minor)
    self.__update_unpacker_version
    if nat_net_major >= 4 and self.use_multicast == False:
      self.__can_change_bitstream = True
    return offset

  def __unpack_server_response(self, data: bytes, packet_size: int) -> int:
    if packet_size == 4:
      with self.__server_responses_lock:
        self.__server_responses.append(int.from_bytes(data, byteorder='little',  signed=True ))
      return 4
    response, _, _ = data[:256].partition(b'\0')
    if len(response) < 30:
      response = response.decode('utf-8')
      if response.startswith('Bitstream'):
        messageList = response.split(',')
        if len(messageList) > 1 and messageList[0] == 'Bitstream':
          nn_version = messageList[1].split('.')
          with self.__server_info_lock:
            template = asdict(self.__server_info)
          if len(nn_version) > 1:
            template["nat_net_major"] = int(nn_version[0])
            template["nat_net_minor"] = int(nn_version[1])
            with self.__server_info_lock:
              self.__server_info = Server_info(**template)
            self.__update_unpacker_version()
      with self.__server_responses_lock:
        self.__server_responses.append(response)
    return len(response)

  def __unpack_server_message(self, data: bytes, __: int) -> int:
    message, _, _ = data.partition(b'\0')
    with self.__server_messages_lock:
      self.__server_messages.append(str(message, encoding='utf-8'))
    return len(message) + 1

  def __unpack_unrecognized_request(self, _: bytes, packet_size: int) -> int:
    logging.error(f"{NAT_Messages.UNRECOGNIZED_REQUEST} - {packet_size = }")
    return packet_size

  def __unpack_undefined_nat_message(self, _: bytes, packet_size: int) -> int:
    logging.error(f"{NAT_Messages.UNDEFINED} - {packet_size = }")
    return packet_size

  def __process_message(self, data: bytes) -> None:
    offset = 0
    message_id = NAT_Messages(self.__get_message_id(data[offset:(offset:=offset+2)]))
    packet_size = int.from_bytes( data[offset:(offset:=offset+2)], byteorder='little', signed=True)
    if message_id not in self.__mapped:
      return
    self.__mapped[message_id](data[offset:], packet_size)

  def __data_thread_function(self) -> None:
    data = bytes()
    logging.info("Data thread start")
    recv_buffer_size=64*1024
    run = True
    while run:
      with self.__running_lock:
        run = self.__running
      try:
        if self.__data_socket is None: return
        data = self.__data_socket.recv(recv_buffer_size)
      except socket.timeout:
        pass
      except socket.error as msg:
        logging.error(f"Data thread {self.local_ip_address}: {msg}")
        data = bytes()
      if len(data):
        self.__process_message(data)
    logging.info("Data thread stopped")

  def __command_thread_function(self) -> None:
    data = bytes()
    logging.info("Command thread start")
    recv_buffer_size=64*1024
    run = True
    keep_alive = b'\n\x00\x00\x00\x00'
    while run:
      with self.__running_lock:
        run = self.__running
      try:
        with self.__command_socket_lock:
          if self.__command_socket is None: return
          if not self.use_multicast:
            self.__command_socket.sendto(keep_alive, (self.server_address, self.command_port))
          data = self.__command_socket.recv(recv_buffer_size)
      except socket.timeout:
        data = bytes()
      except socket.error as msg:
        logging.error(f"Command thread {self.local_ip_address}: {msg}")
        data = bytes()
      if len(data):
        self.__process_message(data)
    logging.info("Command thread stopped")

  def shutdown(self) -> None:
    logging.info(f"Shuting down client {self.server_address}")
    with self.__running_lock:
      self.__running = False
    with self.__command_socket_lock:
      if self.__command_socket is not None:
        self.__command_socket.close()
    if self.__data_socket is not None:
      self.__data_socket.close()
    self.__command_socket = None
    self.__data_socket = None
    self.__data_thread.join()
    self.__command_thread.join()
    self.__freeze = False

  def __del__(self) -> None:
    self.shutdown()